#include "csg_types.hpp"

#include <algorithm>
#include <cmath>
#include <iostream>
#include <limits>

namespace {

const double EXTERNAL_FLOW = -1.0;

template <class... Ts> struct overloaded : Ts... { using Ts::operator()...; };
template <class... Ts>
// clang-format off
overloaded(Ts...) -> overloaded<Ts...>; // not needed as of C++20
// clang-format on

} // namespace

namespace csg {

double signed_distance(const Cone &, double, double, double);
double signed_distance(const Cube &, double, double, double);
double signed_distance(const Cylinder &, double, double, double);
double signed_distance(const Difference &, double, double, double);
double signed_distance(const Intersection &, double, double, double);
double signed_distance(const Mulmatrix &, double, double, double);
double signed_distance(const Sphere &, double, double, double);
double signed_distance(const Type &, double, double, double);

double signed_distance(const Union &group, double xx, double yy, double zz) {
  auto sdist = -std::numeric_limits<double>::max();
  for (const auto &member : group.objs) {
    auto sd = signed_distance(member, xx, yy, zz);
    sdist = std::max(sdist, sd);
  };

  return sdist;
}

double signed_distance(const Intersection &group, double xx, double yy,
                       double zz) {
  auto sdist = std::numeric_limits<double>::max();
  for (const auto &member : group.objs) {
    auto sd = signed_distance(member, xx, yy, zz);
    sdist = std::min(sdist, sd);
  };
  return sdist;
}

double signed_distance(const Difference &group, double xx, double yy,
                       double zz) {
  auto sdist = signed_distance(*group.first_obj, xx, yy, zz);

  for (const auto &member : group.next_objs.objs) {
    auto sd = signed_distance(member, xx, yy, zz);
    sdist = std::min(sdist, -sd);
  };
  return sdist;
}

double signed_distance(const Mulmatrix &mm, double xx, double yy, double zz) {
  // TODO: Invert non-orthogonal matrices
  auto XX = xx - mm.translation[0];
  auto YY = yy - mm.translation[1];
  auto ZZ = zz - mm.translation[2];
  return signed_distance(
      mm.group,
      mm.rotation[0][0] * XX + mm.rotation[1][0] * YY + mm.rotation[2][0] * ZZ,
      mm.rotation[0][1] * XX + mm.rotation[1][1] * YY + mm.rotation[2][1] * ZZ,
      mm.rotation[0][2] * XX + mm.rotation[1][2] * YY + mm.rotation[2][2] * ZZ);
}

double signed_distance(const Cone &cone, double xx, double yy, double zz) {
  double ZZ = cone.center ? zz + cone.height / 2 : zz;
  double sign_z = (ZZ >= 0 && ZZ <= cone.height) ? -1.0 : 1.0;
  auto dist_z = std::min(std::fabs(ZZ), std::fabs(ZZ - cone.height));

  auto rr = cone.radius1 + (ZZ * (cone.radius2 - cone.radius1) / cone.height);
  auto delta_r = xx * xx + yy * yy - rr * rr;
  double sign_r = delta_r <= 0 ? -1.0 : 1.0;
  auto dist_r = std::fabs(delta_r);

  return EXTERNAL_FLOW * std::max(sign_z * dist_z, sign_r * dist_r);
}

double signed_distance(const Cube &cube, double xx, double yy, double zz) {

  auto [Lx, Ly, Lz] = cube.size;
  auto XX = cube.center ? xx + Lx / 2 : xx;
  double sign_x = (XX >= 0 && XX <= Lx) ? -1.0 : 1.0;
  auto dist_x = std::min(std::fabs(XX), std::fabs(XX - Lx));

  auto YY = cube.center ? yy + Ly / 2 : yy;
  double sign_y = (YY >= 0 && YY <= Ly) ? -1.0 : 1.0;
  auto dist_y = std::min(std::fabs(YY), std::fabs(YY - Ly));

  auto ZZ = cube.center ? zz + Lz / 2 : zz;
  double sign_z = (ZZ >= 0 && ZZ <= Lz) ? -1.0 : 1.0;
  auto dist_z = std::min(std::fabs(ZZ), std::fabs(ZZ - Lz));

  return EXTERNAL_FLOW *
         std::max({sign_x * dist_x, sign_y * dist_y, sign_z * dist_z});
}

double signed_distance(const Sphere &sph, double xx, double yy, double zz) {
  auto delta_r = xx * xx + yy * yy + zz * zz - sph.radius * sph.radius;
  double sign = delta_r <= 0 ? -1.0 : 1.0;
  auto dist = std::fabs(delta_r);
  return EXTERNAL_FLOW * sign * dist;
}

double signed_distance(const Cylinder &cyl, double xx, double yy, double zz) {
  auto ZZ = cyl.center ? zz + cyl.height / 2 : zz;
  double sign_z = (ZZ >= 0 && ZZ <= cyl.height) ? -1.0 : 1.0;
  auto dist_z = std::min(std::fabs(ZZ), std::fabs(ZZ - cyl.height));

  auto rr = cyl.radius;
  auto delta_r = xx * xx + yy * yy - rr * rr;
  double sign_r = delta_r <= 0 ? -1.0 : 1.0;
  auto dist_r = std::fabs(delta_r);

  return EXTERNAL_FLOW * std::max(sign_z * dist_z, sign_r * dist_r);
}

double signed_distance(const Type &obj, double xx, double yy, double zz) {

  return std::visit(
      overloaded{
          [xx, yy, zz](auto &&arg) { return signed_distance(arg, xx, yy, zz); },
      },
      obj);
}

} // namespace csg
