#include <memory>

#include "catch2/catch.hpp"

#include <csg.hpp>
#include <csg_types.hpp>

namespace {

TEST_CASE("Translation", "[Levelset Transform]") {

  csg::Cylinder my_cyl{.radius = 2, .height = 20, .center = true};

  auto my_mat = csg::Mulmatrix();
  my_mat.rotation[0] = std::array<double, 3>{{1, 0, 0}};
  my_mat.rotation[1] = std::array<double, 3>{{0, 1, 0}};
  my_mat.rotation[2] = std::array<double, 3>{{0, 0, 1}};
  my_mat.translation = {100, 200, 300};
  my_mat.group.objs.push_back(my_cyl);

  auto my_tree = std::make_shared<csg::Tree>();
  my_tree->top.objs.push_back(my_mat);
  csg::CsgIF my_levelset(my_tree);

  SECTION("Outside") {
    CHECK_FALSE(0 < my_levelset(100, 200, 280));
    CHECK_FALSE(0 < my_levelset(100, 200, 320));
    CHECK_FALSE(0 < my_levelset(103, 200, 300));
    CHECK_FALSE(0 < my_levelset(97, 200, 300));
    CHECK_FALSE(0 < my_levelset(100, 203, 300));
    CHECK_FALSE(0 < my_levelset(100, 197, 300));
  }

  SECTION("Inside") {
    CHECK(0 < my_levelset(100, 200, 295));
    CHECK(0 < my_levelset(100, 200, 305));
    CHECK(0 < my_levelset(101, 200, 300));
    CHECK(0 < my_levelset(99, 200, 300));
    CHECK(0 < my_levelset(100, 201, 300));
    CHECK(0 < my_levelset(100, 199, 300));
  }
}

TEST_CASE("90° Rotation around y-axis, rotating z-axis into x-axis",
          "[Levelset Transform]") {

  csg::Cylinder my_cyl{.radius = 2, .height = 20, .center = false};

  auto my_mat = csg::Mulmatrix();
  my_mat.rotation[0] = std::array<double, 3>{{0, 0, 1}};
  my_mat.rotation[1] = std::array<double, 3>{{0, 1, 0}};
  my_mat.rotation[2] = std::array<double, 3>{{-1, 0, 0}};
  my_mat.translation = {0, 0, 0};
  my_mat.group.objs.push_back(my_cyl);

  auto my_tree = std::make_shared<csg::Tree>();
  my_tree->top.objs.push_back(my_mat);
  csg::CsgIF my_levelset(my_tree);

  SECTION("near origin end") {
    CHECK_FALSE(0 < my_levelset(-1, 0, 0));
    CHECK(0 < my_levelset(1, 0, 0));

    CHECK_FALSE(0 < my_levelset(3, 0, 2.1));
    CHECK(0 < my_levelset(3, 0, 1.9));

    CHECK_FALSE(0 < my_levelset(3, 0, -2.1));
    CHECK(0 < my_levelset(3, 0, -1.9));
  }
  SECTION("near middle") {
    CHECK_FALSE(0 < my_levelset(10, 2.1, 0));
    CHECK(0 < my_levelset(10, 1.9, 0));

    CHECK_FALSE(0 < my_levelset(10, -2.1, 0));
    CHECK(0 < my_levelset(10, -1.9, 0));
  }
  SECTION("near other end") {
    CHECK_FALSE(0 < my_levelset(21, 0, 0));
    CHECK(0 < my_levelset(19, 0, 0));

    CHECK_FALSE(0 < my_levelset(18, 0, 2.1));
    CHECK(0 < my_levelset(18, 0, 1.9));

    CHECK_FALSE(0 < my_levelset(18, 0, -2.1));
    CHECK(0 < my_levelset(18, 0, -1.9));
  }
}

TEST_CASE("Orthogonal rotation + translation of cylinder",
          "[Levelset Transform]") {

  double radius = 4.5, height = 10;
  double Cx = 10, Cy = 5, Cz = 5;

  csg::Cylinder my_cyl{.radius = radius, .height = height, .center = true};
  auto my_mat = csg::Mulmatrix();
  my_mat.rotation[0] = std::array<double, 3>{{0, 0, 1}};
  my_mat.rotation[1] = std::array<double, 3>{{0, 1, 0}};
  my_mat.rotation[2] = std::array<double, 3>{{-1, 0, 0}};
  my_mat.translation = {Cx, Cy, Cz};
  my_mat.group.objs.push_back(my_cyl);

  auto my_tree = std::make_shared<csg::Tree>();
  my_tree->top.objs.push_back(my_mat);
  csg::CsgIF my_levelset(my_tree);

  SECTION("near origin end") {
    CHECK_FALSE(0 < my_levelset(Cx - (1.1 * height / 2), Cy, Cz));
    CHECK(0 < my_levelset(Cx - (0.9 * height / 2), Cy, Cz));

    CHECK_FALSE(0 <
                my_levelset(Cx - (0.9 * height / 2), Cy, Cz + 1.1 * radius));
    CHECK(0 < my_levelset(Cx - (0.9 * height / 2), Cy, 0.9 * radius));

    CHECK_FALSE(0 <
                my_levelset(Cx - (0.9 * height / 2), Cy, Cz - 1.1 * radius));
    CHECK(0 < my_levelset(Cx - (0.9 * height / 2), Cy, Cz - 0.9 * radius));
  }
  SECTION("near middle") {
    CHECK_FALSE(0 < my_levelset(Cx, Cy + 1.1 * radius, Cz));
    CHECK(0 < my_levelset(Cx, Cy + 0.9 * radius, Cz));

    CHECK_FALSE(0 < my_levelset(Cx, Cy - 1.1 * radius, Cz));
    CHECK(0 < my_levelset(Cx, Cy - 0.9 * radius, Cz));
  }
  SECTION("near other end") {
    CHECK_FALSE(0 < my_levelset(Cx + (1.1 * height / 2), Cy, Cz));
    CHECK(0 < my_levelset(Cx + (0.9 * height / 2), Cy, Cz));

    CHECK_FALSE(0 <
                my_levelset(Cx + (0.9 * height / 2), Cy, Cz + 1.1 * radius));
    CHECK(0 < my_levelset(Cx + (0.9 * height / 2), Cy, Cz + 0.9 * radius));

    CHECK_FALSE(0 <
                my_levelset(Cx + (0.9 * height / 2), Cy, Cz - 1.1 * radius));
    CHECK(0 < my_levelset(Cx + (0.9 * height / 2), Cy, Cz - 0.9 * radius));
  }
}

} // namespace
